/* $LAAS: viamMainCodels.c 2008/08/04 16:32:41 mallet $ */

/*
 * Copyright (c) 2007-2008 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 *                                      Anthony Mallet on Tue Jan 23 2007
 */

#include <portLib.h>
#include <stdio.h>
#include <sys/time.h>

#include <viam/viamlib.h>

#include "server/viamHeader.h"
#include "shared.h"


/* --- local data ---------------------------------------------------- */

viam_handle_t vh;		/**< viam handle */


/* --- viamMainInit -------------------------------------------------- */

/** \brief Initialization codel (fIDS, ...)
*
 * \return OK or ERROR
 */

STATUS
viamMainInit(int *report)
{
  SDI_F->init_done = 0;
  vh = viam_init();

#ifdef VIAM_DISPLAY
  if (viamDisplayInit(report) != OK) return ERROR;
#endif
  return viamAcqInit(report);
}


/* --- viamMainEnd --------------------------------------------------- */

/** \brief Termination codel of Main execution task
 *
 * \return OK or ERROR
 */

STATUS
viamMainEnd(void)
{
  SDI_F->init_done = 0;

  viamAcqEnd();
  viam_hardware_detach(vh);
  viam_release(vh);
  return OK;
}


/* --- Init ---------------------------------------------------------- */

/** \brief Look for camera nodes and configure module.
 *
 * Reports:      OK
 *              S_viam_IEEE1394_CREATE_HANDLE
 */

ACTIVITY_EVENT
viamInitMain(int *report)
{
  viam_hardware_attach(vh);

  SDI_F->init_done = 1;
  return ETHER;
}


/* --- Configure ----------------------------------------------------- */

/** \brief Configure a bank for image acquisition
 *
 * codel EXEC of Configure
 *
 * \return EXEC END ETHER FAIL ZOMBIE
 */

ACTIVITY_EVENT
viamConfigureMain(ViamId *name, int *report)
{
  viam_bankdata_t *d;
  viam_bank_t bank;
  STATUS s;

  bank = viam_bank_getbyname(vh, name->id);
  if (!bank) VIAM_RETURN_WITHERRNO(report, ETHER);
  d = viam_bank_usrdata(bank);
  if (d->pdata) { *report = S_viam_CONFIGURED; return ETHER; }

  /* connect to pom */
  s = viamPomInit(bank, report);
  if (s != OK) return ETHER;

  /* configure hardware */
  if (viam_bank_configure(vh, bank))
    VIAM_RETURN_WITHERRNO(report, ETHER);

  /* initialize poster */
  d->pdata = viamPosterCreate(vh, bank, d->buffers, report);
  if (!d->pdata) {
    viam_bank_unconfigure(vh, bank);
    return ETHER;
  }

  return ETHER;
}


/* --- Reset --------------------------------------------------------- */

/** \brief Stop module and reset configuration info
 *
 * codel EXEC of Reset
 *
 * \returns  EXEC END ETHER FAIL ZOMBIE
 */

ACTIVITY_EVENT
viamResetMain(int *report)
{
  viamMainEnd();
  viamMainInit(report);
  return ETHER;
}


/* --- FilterGetCap -------------------------------------------------- */

/** \brief EXEC codel of FilterGetCap
 *
 * Reports:      OK
 *              S_viam_NO_SUCH_CAMERA
 *              S_viam_FWBUS
 *              S_viam_ARRAY_SIZE
 *
 * \return ETHER
 */

ACTIVITY_EVENT
viamFilterGetCapMain(ViamFilterGetCap *getcap, ViamFilterCap *cap, int *report)
{
  viam_camera_t camera;
  unsigned int methods, automodes;
  unsigned int i, count;

  camera = viam_camera_getbyname(vh, getcap->camera.id);
  if (!camera) VIAM_RETURN_WITHERRNO(report, ETHER);

  if (viam_queryfeature(vh, camera, getcap->type, &methods, &automodes))
    VIAM_RETURN_WITHERRNO(report, ETHER);

  memset(cap, 0, sizeof(*cap));
  count = 0;
  for(i=1; i<0x10; i<<=1) if (methods & i) {
    if (count >= sizeof(cap->methods)/sizeof(cap->methods[0])) {
      *report = S_viam_ARRAY_SIZE;
      return ETHER;
    }

    cap->methods[count++] = i;
  }
  count = 0;
  for(i=1; i<0x10; i<<=1) if (automodes & i) {
    if (count >= sizeof(cap->automodes)/sizeof(cap->automodes[0])) {
      *report = S_viam_ARRAY_SIZE;
      return ETHER;
    }

    cap->automodes[count++] = i;
  }

  return ETHER;
}


/* --- viamCameraListModesMain --------------------------------------- */

/** \brief EXEC codel of CameraListModes
 *
 * List available acquisition modes
 * Reports: OK, S_viam_FWBUS
 *
 * \return ETHER
 */

ACTIVITY_EVENT
viamCameraListHWModesMain(ViamId *name, ViamHWModeList *hwmodes, int *report)
{
  viam_camera_t camera;
  int i;

  camera = viam_camera_getbyname(vh, name->id);
  if (!camera) VIAM_RETURN_WITHERRNO(report, ETHER);

  hwmodes->count =
    viam_queryhwmodes(vh, camera, hwmodes->modes,
		      sizeof(hwmodes->modes)/sizeof(hwmodes->modes[0]));

  /* reset extra elements */
  for(i = hwmodes->count;
      i<sizeof(hwmodes->modes)/sizeof(hwmodes->modes[0]); i++) {
    hwmodes->modes[i].size = VIAM_HWSZ_INVALID;
    hwmodes->modes[i].format = VIAM_HWFMT_INVALID;
    hwmodes->modes[i].crop = VIAM_HW_FIXED;
    hwmodes->modes[i].fps = VIAM_HWFPS_INVALID;
  }

  return ETHER;
}
