/* $LAAS: viamPoster.c 2010/06/02 11:06:05 mallet $ */

/*
 * Copyright (c) 2007-2010 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 *                                      Anthony Mallet on Mon Mar  5 2007
 */

#include <portLib.h>
#include <time.h>
#include <err.h>

#include <viam/viamlib.h>
#include <pom/libeuler/pomEuler.h>

#include "server/viamHeader.h"
#include "shared.h"


/* --- local data ---------------------------------------------------- */

#define MAX_BUFFERS	2

struct viam_posterimagedata_t {
  IplImage *image;			/**< image header */
  unsigned long offset[MAX_BUFFERS];	/**< image buffers offsets */
  int lastindex;			/**< last buffer used */
};

struct viam_posterdata_t {
  POSTER_ID posterId;
  ViamImageBank *poster;
  int buffers;				/*<< number of buffers */
  struct viam_posterimagedata_t *idata;	/*<< array for each camera */
};


/* --- viamPosterCreate ---------------------------------------------- */

/** Create image poster.
 *
 * \param[in]	vh		Viam handle.
 * \param[in]	bank		Bank object.
 * \param[out]	report		Error code.
 */

viam_posterdata_t *
viamPosterCreate(viam_handle_t vh, viam_bank_t bank, int buffers, int *report)
{
  viam_posterdata_t *pdata;
  struct viam_posterimagedata_t *idata;
  int nnodes;
  ViamImageHeader *h;
  viam_image_ref c;
  viam_geometry_t geom;
  int i, j, s;
  size_t size;
  char *name;
  STATUS e;

  if (buffers > MAX_BUFFERS) { *report = S_viam_INVALID_VALUE; return NULL; }
  nnodes = viam_bank_nimages(bank);

  /* allocate local data */
  pdata = malloc(sizeof(*pdata));
  if (!pdata) { *report = S_viam_MEMORY; return NULL; }
  pdata->poster = NULL;
  pdata->posterId = NULL;
  pdata->buffers = buffers;
  pdata->idata = malloc(nnodes * sizeof(*pdata->idata));
  if (!pdata->idata) { *report = S_viam_MEMORY; goto error; }


  /* compute poster size */
  size = (char *)&((ViamImageBank *)NULL)->image[nnodes] - (char *)NULL;
  for(i=0; i<nnodes; i++) {
    s = viam_image_geometry(vh, viam_bank_image(vh, bank, i), &geom);
    if (s) { VIAM_MAPERRNO(report); goto error; }

    /* multiply the actual size by a factor `buffers' for double buffering */
    size += buffers *
      geom.width * geom.height * geom.channels * (((geom.depth & 0xFF) + 7) >> 3);
  }

  /* allocate shared memory for poster */
  name = alloca(strlen("viam") + strlen(viam_bank_name(bank) + 1));
  sprintf(name, "%s%s", "viam", viam_bank_name(bank));

  e = posterCreate(name, size, &pdata->posterId);
  if (e != OK) { *report = errnoGet(); goto error; }
  pdata->poster = posterAddr(pdata->posterId);

  /* initialize poster data */
  e = posterTake(pdata->posterId, POSTER_WRITE);
  if (e != OK) { *report = errnoGet(); goto error; }

  snprintf(pdata->poster->name.id, sizeof(pdata->poster->name.id),
	   "%s", viam_bank_name(bank));
  pdata->poster->nImages = nnodes;

  /* XXX temporary put all the calibration data here, until I rewrite
   * the code to put only what is needed */
  {
    s = viam_bank_calibration(vh, bank, &pdata->poster->calibration);
  }

  size = (char *)&((ViamImageBank *)NULL)->image[nnodes] - (char *)NULL;
  for(i=0; i<nnodes; i++) {
    c = viam_bank_image(vh, bank, i);
    idata = &pdata->idata[i];
    h = &pdata->poster->image[i];
    snprintf(h->name.id, sizeof(h->name.id), "%s", viam_image_name(c));

    s = viam_image_geometry(vh, c, &geom);
    if (s) { VIAM_MAPERRNO(report); goto error; }

    /* image header */
    h->tacq_sec = 0;
    h->tacq_usec = 0;

    h->nChannels = geom.channels;
    h->depth = geom.depth;
    h->width = geom.width;
    h->height = geom.height;
    h->widthStep =
      geom.width * geom.channels * (((geom.depth & 0xFF) + 7) >> 3);
    h->imageSize = h->widthStep * h->height;

    /* XXX temporary put all the calibration data here, until I rewrite
     * the code to put only what is needed */
    {
      s = viam_image_calibration(vh, c, &h->calibration, &geom);
    }

    h->dataOffset = size;
    h->dataOffset -= (char *)((ViamImageBank *)NULL)->image[i].data - (char *)NULL;
    memset(&h->data[h->dataOffset], 0, h->imageSize);

    /* internal image data */
    idata->image = cvCreateImageHeader(
      cvSize(geom.width, geom.height), geom.depth, geom.channels);
    if (!idata->image) { *report = S_viam_MEMORY; goto error; }
    cvSetData(idata->image, &h->data[h->dataOffset], h->widthStep);
    for(j = 0; j < buffers; j++)
      idata->offset[j] = h->dataOffset + j * h->imageSize;
    idata->lastindex = 0;

    /* multiply the actual size by a factor `buffers' for double buffering */
    size += buffers * h->imageSize;
  }

  posterGive(pdata->posterId);
  return pdata;

error:
  if (pdata) {
    if (pdata->posterId) posterDelete(pdata->posterId);
    if (pdata->idata) free(pdata->idata);
    free(pdata);
  }
  return NULL;
}


/* --- viamPosterDelete ---------------------------------------------- */

/** Delete image poster.
 *
 * \param[in]	vh	Viam handle.
 * \param[in]	bank
 */

STATUS
viamPosterDelete(viam_handle_t vh, viam_posterdata_t *pdata)
{
  return OK;
}


/* --- viamPosterAcqLock --------------------------------------------- */

/** \brief Lock poster for image acquisition.
 *
 * For simple buffering, the fuction locks the poster for writing. For
 * double buffering, the function does nothing.
 *
 * \param[in] vh	Viam handle.
 * \param[in] pdata	Poster data.
 * \param[out] report	Error code in case of failure.
 */

STATUS
viamPosterAcqLock(viam_handle_t vh, viam_posterdata_t *pdata, int *report)
{
  if (pdata->buffers < 2) {
    STATUS e;

    e = posterTake(pdata->posterId, POSTER_WRITE);
    if (e != OK) { *report = errnoGet(); return ERROR; }
  }

  return OK;
}


/* --- viamPosterAcqRelease ------------------------------------------ */

/** \brief Unlock poster after image acquisition
 *
 * For simple buffering, the fuction unlocks the poster. For double
 * buffering, the function updates the image data offsets in the poster
 * structure. The poster is locked for writing during this operation.
 *
 * \param[in] vh	Viam handle.
 * \param[in] pdata	Poster data.
 * \param[out] report	Error code in case of failure.
 */

STATUS
viamPosterAcqRelease(viam_handle_t vh, viam_bank_t bank,
		     viam_posterdata_t *pdata, int *report)
{
  int nnodes = viam_bank_nimages(bank);
  int i;

  /* update image */
  if (pdata->buffers > 1) {
    struct viam_posterimagedata_t *idata;
    STATUS e;
    int i;

    e = posterTake(pdata->posterId, POSTER_WRITE);
    if (e != OK) { *report = errnoGet(); return ERROR; }

    for(i=0; i<pdata->poster->nImages; i++) {
      idata = &pdata->idata[i];
      pdata->poster->image[i].dataOffset = idata->offset[idata->lastindex];
    }
  }

  /* update position and time of acquisition */
  for(i=0; i<nnodes; i++) {
    viam_imagedata_t *idata = viamIData(bank, i);

    pdata->poster->image[i].tacq_sec = idata->tacq.tv_sec;
    pdata->poster->image[i].tacq_usec = idata->tacq.tv_usec;
    pdata->poster->image[i].pos = idata->pomFrame;
  }

  posterGive(pdata->posterId);

#ifdef VIAM_DISPLAY
  for(i=0; i<nnodes; i++) {
    viam_imagedata_t *idata = viamIData(bank, i);

    /* performance tag */
    if (idata->vtags == VIAM_ON) {
      idata->fps = .9 * idata->fps + .1 / (
	(idata->tacq.tv_sec + idata->tacq.tv_usec * 1e-6) -
	(idata->ptacq.tv_sec + idata->ptacq.tv_usec * 1e-6)
	);
      idata->ptacq = idata->tacq;
    }
  }
  viamDisplaySignal(bank);
#endif

  /* save */
  for(i=0; i<nnodes; i++) if (viamIData(bank, i)->file) {
    viam_bankdata_t *d = viam_bank_usrdata(bank);
    viam_imagedata_t *idata = viamIData(bank, i);
    char path[1024];
    char *ext;
    FILE *f;

    snprintf(path, sizeof(path), idata->file, idata->seq);
    cvSaveImage(path, pdata->idata[i].image
#if CV_MAJOR_VERSION > 1
		, NULL
#endif
      );

    /* save the time of acquisition */
    ext = strrchr(path, '.');
    if (ext) strcpy(ext, ".time"); else strcat(path, ".time");
    f = fopen(path, "w");
    if (f) {
      fprintf(f, "%d.%06d\n", idata->tacq.tv_sec, idata->tacq.tv_usec);
      fclose(f);
    }

    /* save the camera position if POM is enabled */
    if (d->tags == VIAM_ENABLE) {
      /* get the dir where the images are saved */
      ext = strrchr(path, '.');
      if (ext) strcpy(ext, ".pos"); else strcat(path, ".pos");
      pomWriteSensorPos(path, &idata->pomFrame);
    }

    idata->seq++;
  }

  return OK;
}


/* --- viamPosterNewImage -------------------------------------------- */

/** Return image structure suitable for a new acquisition
 */

IplImage *
viamPosterNewImage(viam_handle_t vh, viam_posterdata_t *pdata, int i)
{
  struct viam_posterimagedata_t *idata = &pdata->idata[i];

  if (pdata->buffers > 1) {
    ViamImageHeader *h = &pdata->poster->image[i];

    idata->lastindex++;
    if (idata->lastindex >= pdata->buffers) idata->lastindex = 0;
    cvSetData(idata->image,
	      &h->data[idata->offset[idata->lastindex]], h->widthStep);
  }

  return idata->image;
}


/* --- viamPosterCurrentImage ---------------------------------------- */

/** Return image structure currently used
 */

IplImage *
viamPosterCurrentImage(viam_handle_t vh, viam_posterdata_t *pdata, int i)
{
  return pdata->idata[i].image;
}
